/****************************************************************************/
// Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
// Copyright (C) 2012-2022 German Aerospace Center (DLR) and others.
// This program and the accompanying materials are made available under the
// terms of the Eclipse Public License 2.0 which is available at
// https://www.eclipse.org/legal/epl-2.0/
// This Source Code may also be made available under the following Secondary
// Licenses when the conditions for such availability set forth in the Eclipse
// Public License 2.0 are satisfied: GNU General Public License, version 2
// or later which is available at
// https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
/****************************************************************************/
/// @file    libsumo.h
/// @author  Michael Behrisch
/// @date    03.03.2021
///
// Helper file for libsumo inclusion
/****************************************************************************/
#pragma once
#include "Edge.h"
#include "GUI.h"
#include "InductionLoop.h"
#include "Junction.h"
#include "LaneArea.h"
#include "Lane.h"
#include "MultiEntryExit.h"
#include "POI.h"
#include "Polygon.h"
#include "Route.h"
#include "Simulation.h"
#include "TrafficLight.h"
#include "VehicleType.h"
#include "Vehicle.h"
#include "Person.h"
#include "Calibrator.h"
#include "BusStop.h"
#include "ParkingArea.h"
#include "ChargingStation.h"
#include "OverheadWire.h"
#include "Rerouter.h"
#include "MeanData.h"
#include "VariableSpeedSign.h"
#include "RouteProbe.h"
