/****************************************************************************/
// Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
// Copyright (C) 2001-2022 German Aerospace Center (DLR) and others.
// This program and the accompanying materials are made available under the
// terms of the Eclipse Public License 2.0 which is available at
// https://www.eclipse.org/legal/epl-2.0/
// This Source Code may also be made available under the following Secondary
// Licenses when the conditions for such availability set forth in the Eclipse
// Public License 2.0 are satisfied: GNU General Public License, version 2
// or later which is available at
// https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
/****************************************************************************/
/// @file    flag_plus.cpp
/// @author  Daniel Krajzewicz
/// @author  Michael Behrisch
/// @date    2004-11-23
///
// Generated by reswrap version 3.0.0
// created by reswrap from file flag_plus.gif
/****************************************************************************/


const unsigned char flag_plus[] = {
    0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x10, 0x00, 0x10, 0x00, 0xa2, 0xff, 0x00, 0xc0, 0xc0, 0xc0,
    0xff, 0xc0, 0x00, 0xff, 0x00, 0x00, 0x12, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x21, 0xf9, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x2c, 0x00, 0x00,
    0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x40, 0x03, 0x3d, 0x08, 0x40, 0xa4, 0xde, 0x2e, 0xae, 0x00,
    0xc9, 0x18, 0x22, 0x8b, 0x5a, 0x69, 0x86, 0x0f, 0x25, 0x2d, 0x16, 0xc7, 0x30, 0x13, 0x3a, 0x2a,
    0x04, 0x75, 0x61, 0x9b, 0xd4, 0x82, 0xd6, 0x3b, 0x70, 0x6b, 0xae, 0xb2, 0x62, 0x74, 0xfe, 0x9e,
    0x98, 0xac, 0x31, 0xbb, 0x08, 0x47, 0xb3, 0x9a, 0x66, 0x37, 0x7b, 0xd8, 0x3e, 0x29, 0xd9, 0xab,
    0x03, 0x42, 0xe6, 0xae, 0x8a, 0x04, 0x00, 0x3b
};

