/****************************************************************************/
// Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
// Copyright (C) 2001-2022 German Aerospace Center (DLR) and others.
// This program and the accompanying materials are made available under the
// terms of the Eclipse Public License 2.0 which is available at
// https://www.eclipse.org/legal/epl-2.0/
// This Source Code may also be made available under the following Secondary
// Licenses when the conditions for such availability set forth in the Eclipse
// Public License 2.0 are satisfied: GNU General Public License, version 2
// or later which is available at
// https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
/****************************************************************************/
/// @file    manip.cpp
/// @author  Daniel Krajzewicz
/// @author  Michael Behrisch
/// @date    2004-11-23
///
// Generated by reswrap version 3.0.0
// created by reswrap from file manip.gif
/****************************************************************************/


const unsigned char manip[] = {
    0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x10, 0x00, 0x10, 0x00, 0xa2, 0xff, 0x00, 0xc0, 0xc0, 0xc0,
    0xf2, 0xe8, 0x3a, 0xef, 0xc0, 0x01, 0xa0, 0xa5, 0xc7, 0x9e, 0x66, 0x0d, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x21, 0xf9, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x2c, 0x00, 0x00,
    0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x40, 0x03, 0x30, 0x08, 0xba, 0xdc, 0x5d, 0x70, 0x14, 0x47,
    0x59, 0x91, 0x10, 0x94, 0x20, 0x48, 0x9f, 0x55, 0xa8, 0x5c, 0x52, 0x78, 0x81, 0x0f, 0xf7, 0x55,
    0x45, 0x47, 0xa0, 0xa2, 0x99, 0xc5, 0xda, 0x50, 0x9a, 0x37, 0x0e, 0xa7, 0x82, 0xb0, 0x5b, 0x1c,
    0xcf, 0x8b, 0xd5, 0x5b, 0x89, 0x20, 0x3f, 0x47, 0x02, 0x00, 0x3b
};

