/****************************************************************************/
// Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
// Copyright (C) 2001-2022 German Aerospace Center (DLR) and others.
// This program and the accompanying materials are made available under the
// terms of the Eclipse Public License 2.0 which is available at
// https://www.eclipse.org/legal/epl-2.0/
// This Source Code may also be made available under the following Secondary
// Licenses when the conditions for such availability set forth in the Eclipse
// Public License 2.0 are satisfied: GNU General Public License, version 2
// or later which is available at
// https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
/****************************************************************************/
/// @file    flag_minus.cpp
/// @author  Daniel Krajzewicz
/// @author  Michael Behrisch
/// @date    2004-11-23
///
// Generated by reswrap version 3.0.0
// created by reswrap from file flag_minus.gif
/****************************************************************************/

const unsigned char flag_minus[] = {
    0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x10, 0x00, 0x10, 0x00, 0xa2, 0xff, 0x00, 0xc0, 0xc0, 0xc0,
    0xff, 0xc0, 0x00, 0xff, 0x00, 0x00, 0x12, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x21, 0xf9, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x2c, 0x00, 0x00,
    0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x40, 0x03, 0x3b, 0x08, 0xba, 0x4b, 0xfc, 0x24, 0x38, 0x40,
    0xc6, 0x10, 0x58, 0x4c, 0x16, 0x09, 0xde, 0x4a, 0xf7, 0x50, 0xd5, 0x44, 0x9c, 0x28, 0x25, 0x81,
    0x90, 0x64, 0x5d, 0x5a, 0xbb, 0x55, 0xef, 0x60, 0xb2, 0x63, 0xde, 0x48, 0x10, 0x8a, 0x4a, 0x1f,
    0x9d, 0x8a, 0x16, 0xcb, 0x75, 0x88, 0x9f, 0x19, 0x2f, 0x54, 0x0b, 0x8a, 0x38, 0x2f, 0xd3, 0xd2,
    0x28, 0xac, 0x26, 0x00, 0x00, 0x3b
};

